#include <sys/times.h>
#include <iostream>
#include <fstream>
#include <string>
#include <math.h>

#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_cdf.h>

#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/boyer_myrvold_planar_test.hpp>

using namespace boost;

typedef boost::adjacency_list<boost::vecS, boost::vecS, boost::undirectedS, boost::property<vertex_index_t, int> > Graph;
typedef graph_traits<Graph>::edge_iterator EdgeIt;
typedef graph_traits<Graph>::out_edge_iterator OutEdgeIt;
typedef graph_traits<Graph>::vertex_iterator VertexIt;
typedef graph_traits<Graph>::edge_descriptor EdgeD;

template<typename Graph, typename EdgeIt, typename VertexIt>
void g2Dot(Graph g, std::string filename)
{
  std::ofstream fout(filename.c_str());
  fout << "graph G {" << std::endl;
  fout << "graph[size=\"100,100\"];" << std::endl;

  // var nodes 
  fout << "node[shape=circle,width=0.25,height=0.25,fixedsize=true];" << std::endl;
	VertexIt vi, ve;
	for (tie(vi, ve) = vertices(g); vi != ve; vi++) {
    fout << "\t" << *vi;
    fout << " [style=\"setlinewidth(1)\",label=" << *vi << "];" << std::endl;
  }

	// edges 
  EdgeIt ei, ei_end;
  for (tie(ei, ei_end) = edges(g); ei != ei_end; ei++) {
		fout << "\t" << source(*ei, g) << " -- " << target(*ei, g);
		fout << "[style=\"setlinewidth(1)\",fontsize=15,label=\"";
		fout << "\"];" << std::endl;
  }
  fout << "}" << std::endl;
  fout.close();
}

void writeFg(const Graph &g, std::string filename, gsl_rng *r, double sigma, double theta)
{
	std::ofstream fout( filename.c_str() );
	fout << num_edges(g) + num_vertices(g) << std::endl << std::endl;// + num_vertices(g)

	// pairwise potentials
	VertexIt vb, ve;
	for (tie(vb, ve) = vertices(g); vb != ve; vb++) {
		OutEdgeIt eb, ee;
		for (tie(eb, ee) = out_edges(*vb, g); eb != ee; eb++) {

			if (*vb < target(*eb, g)) {

				fout << 2 << std::endl;
				fout << *vb << " " << target(*eb, g) << std::endl;
				fout << "2 2" << std::endl<< 4 << std::endl;
		
				double w = gsl_ran_gaussian(r, sigma);
		
				fout << "0\t" << exp(w) << std::endl;
				fout << "1\t" << exp(-w) << std::endl;
				fout << "2\t" << exp(-w) << std::endl;
				fout << "3\t" << exp(w) << std::endl << std::endl;
			}
		}

		// single variable potential
		fout << 1 << std::endl;
		fout << *vb << std::endl;
		fout << "2" << std::endl<< 2 << std::endl;
		
		double w = gsl_ran_gaussian(r, theta);

		fout << "0\t" << exp(w) << std::endl;
		fout << "1\t" << exp(-w) << std::endl << std::endl;
	}
	fout.close();
}

int main( int argc, char *argv[] ) {
	int n, m, e;
	double sigma, theta;

	if (argc != 7) {
		std::cerr << "./create_pstar filename num_stars num_nodes_star num_random_edges theta sigma" << std::endl;
		exit(-1);
	}

	std::string sfilename(argv[1]);
	n = atoi(argv[2]);
	m = atoi(argv[3]);
	e = atoi(argv[4]);
	theta = atof(argv[5]);
	sigma = atof(argv[6]);

	std::cout << "filename         " << sfilename << std::endl;
	std::cout << "num_stars        " << n << std::endl;
	std::cout << "num_nodes_star   " << m << std::endl;
	std::cout << "num_random_edges " << e << std::endl;
	std::cout << "theta            " << theta << std::endl;
	std::cout << "sigma            " << sigma << std::endl;
	
	// Init pseudorandom generator
  gsl_rng_env_setup();
  gsl_rng *r = gsl_rng_alloc(gsl_rng_mrg);
 
	// create stars
	Graph g;
	for (int i=0; i<n; i++) {
		add_vertex(g);
		for (int j=0; j<m; j++) {
			 add_vertex(g);
			 add_edge(i+i*m, i+i*m+j+1, g);
			 if (j) add_edge(i+i*m+j, i+i*m+j+1, g);
		}
		add_edge(i+i*m+m, i+i*m+1, g);
		if (i) add_edge(i-1+(i-1)*m, i+i*m, g);
	}
	int N = num_vertices(g);

	if (boyer_myrvold_planarity_test(g)) {
	
		// write .fg graph
		char buf[100];
		sprintf(buf, "%s_0.fg", sfilename.c_str());
		std::cout << "creating " << buf << std::endl;
		writeFg(g, std::string(buf), r, sigma, theta);
	
		// write .dot graph
		sprintf(buf, "%s_0.dot", sfilename.c_str());
		std::cout << "creating " << buf << std::endl;
		g2Dot<Graph, EdgeIt, VertexIt>(g, std::string(buf) );
	}
	else { std::cout << "Initial graph is not planar..." << std::endl; exit(-1); }

	// connect stars preserving planarity
	int tries = 0, tries2 = 0, e1 = 1, u, v; 
	while ((e1 <= e) && (tries < 100)) {

		Graph g1 = g;

		// pick up 2 random vertices
		bool fi = false;
		while (!fi && (tries2 < 100)) {
			u = gsl_rng_uniform_int(r, N);
			v = gsl_rng_uniform_int(r, N);
			std::pair<EdgeD, bool> ed = edge(u, v, g);
			if ((!ed.second)&&(u!=v)) {
				fi = true;
				tries2 = 0;
			}
		}
		if (tries2 == 100) {
			std::cout << "Cannot add more edges..." << std::endl;
			break;
		}

		add_edge(u,v, g1);

		// check if planar
	  if (boyer_myrvold_planarity_test(g1)) {
	    std::cout << "g is planar." << std::endl;
			add_edge(u,v, g);
			tries = 0;
			e1++;

			// write .dot graph
			char buf[100];
			sprintf(buf, "%s_%d.dot", sfilename.c_str(), e1);
			std::cout << "creating " << buf << std::endl;
			g2Dot<Graph, EdgeIt, VertexIt>(g1, std::string(buf) );

			// write .fg graph
			sprintf(buf, "%s_%d.fg", sfilename.c_str(), e1);
			writeFg(g, std::string(buf), r, sigma, theta);
		}
	  else {
	    std::cout << "g is not planar!" << std::endl;
			tries++;
		}
												
	}

	if (tries == 100) {
		std::cout << "Cannot add more edges..." << std::endl;
	}

	std::ofstream fout(sfilename.c_str());
	fout << 3*n+1 << std::endl << std::endl;
	

  return 0;
}
