#ifndef __GRAPH_H__
#define __GRAPH_H__

//#define DEBUG

// TODO: operator < should be invisible, specify a compare function to
// set<Node> and set<Clique> instead, need to alter myset.h

#include <string>
#include <iostream>
#include <sstream>
#include <list>
#include "myset.h"
#include "factor.h"

//#define DEBUG

/*	These macro's are used to define a lot of member functions for simple
 *	creation of cliques.  Example: graph.clique(a,b, 0.1,0.3,0.2,0.4) inserts
 *	a clique for nodes a and b with probability table [0.1,0.3,0.2,0.4]
 */

#define FUNC_SHORT(n) \
	Clique clique (XARG##n(Node), const double *p) \
		{ return(insert(_Clique(XCALL##n, p))); };
#define FUNC_LONG(n) \
	Clique clique (XARG##n(Node), PARG##n) \
		{ return(insert(_Clique(XCALL##n, PCALL##n))); };

#define CLIQUE_FUNCS \
		FUNC_SHORT(1) FUNC_SHORT(2) FUNC_SHORT(3) FUNC_SHORT(4) \
		FUNC_SHORT(5) FUNC_SHORT(6) \
		FUNC_LONG(1) FUNC_LONG(2) FUNC_LONG(3) FUNC_LONG(4)

using namespace std;
// this should be "namespace std {", but I get some compile errors

class Graph {
public:

/*	These are the main classes defined in- and outside Graph:
 * 		Node: An iterator into the private _NodeList _nodes.
 *		NodeSet: A set of Nodes (thus _NodeList iterators).
 *		Clique: An iterator into the private _CliqueList _cliques.
 *			Clique inherits from factor<Node>, but can be seen as a NodeSet
 *			which holds a probability table as well.
 *		CliqueSet: A set of Cliques (this _CliqueList iterators).
 */ 

	class Node;
	class Clique;
	typedef myset<Node> NodeSet;
	typedef myset<Clique> CliqueSet;

private:

	struct _Node;
	struct _Clique;

	// pass is used in recursive functions to get a new number ++pass that
	// is garanteed to be bigger than all previously used numbers
	// (I assume pass will never cycle, which is reasonable)
	static unsigned long pass;
	static unsigned long node_label;

	struct _Node: public CliqueSet {
		string label;
		NodeSet _neighbors;
		mutable unsigned long pass;
		int _missing_links; // caching variable

		_Node (char *label) {
			this->pass = 0;
			if(label) this->label = label;
			else {
				for(unsigned long x = ++node_label; x; x /= 10)
					this->label += ('0' + x % 10);
			};
		};
	};

	typedef list<_Node> _NodeList;
	_NodeList _nodes;

public:
	class Node: protected _NodeList::iterator {
		friend class Clique;
		friend class Graph;

	private:
		// Nodes do not have a real ordening.  For the NodeSet, however,
		// there need to exist some ordening.  Here it is by physical
		// address of the Node structs in memory.
		//
		// (I can't get this member function private, TODO)
		//bool operator < (Node node) const
		//	{ return(&**this < &*node); };

	public:
		Node (): _NodeList::iterator() {};
		Node (_NodeList::iterator it): _NodeList::iterator(it) {};

		bool operator == (Node node) const
			{ return(&**this == &*node); };
		bool operator != (Node node) const
			{ return(&**this != &*node); };
		bool operator < (Node node) const
			{ return(&**this < &*node); };

		bool connected_with (Node node, unsigned long pass = 0) const;
		bool operator >> (Node node) const
			{ return(connected_with(node)); };

		const CliqueSet& cliques () const { return(**this); };
		const NodeSet& neighbors () const { return((**this)._neighbors); };

		// computes the number of links to add to make its neighborhood a
		// fully connected clique
		int missing_links () const;

		Node& operator ++ () {
			_NodeList::iterator::operator++();
			return(*this);
		};

		pair<logd, logd> marginal () const;
		double marginal_negative () const {
			pair<logd, logd> m = marginal();
			return(m.first / (m.first + m.second));
		};
		double marginal_positive () const {
			pair<logd, logd> m = marginal();
			return(m.second / (m.first + m.second));
		};
		double logNorm () const { 
			pair<logd, logd> m = marginal();
			return(m.first + m.second);
		}

		const string& label () const { return((**this).label); };
		friend ostream& operator << (ostream& os, const Node& node);
	};
private:
	// structure to sort nodes in elimination order
	struct EliminationCmp {
        bool operator () (Node a, Node b) const {
			if(a->_missing_links < b->_missing_links) return(true);
			if(b->_missing_links < a->_missing_links) return(false);
			if(a.neighbors().size() < b.neighbors().size()) return(true);
			if(b.neighbors().size() < a.neighbors().size()) return(false);
			return(&*a < &*b);
		};
	};

private:
	typedef pair<Clique, Clique> CliquePair;

	struct _Clique: public factor<Node> {
		CliqueSet _neighbors;
		mutable unsigned long pass;
		int union_size; // used in separators to store the size of the union
		int number;

		_Clique (const NodeSet& ns, double z): factor<Node>(ns, z)
			{ pass = 0; };
		_Clique (const NodeSet& ns, const double *p): factor<Node>(ns, p)
			{ pass = 0; };
		_Clique (int n, const Node *x, const double *p): factor<Node>(n, x, p)
			{ pass = 0; };
		FACTOR_CONSTRUCTORS(_Clique, Node)

		void set_number (int number, unsigned long pass = 0) {
			if(!pass) pass = ++Graph::pass;
			if(this->pass == pass) return;
			this->pass = pass;
			this->number = number;
			for(CliqueSet::iterator iclique = _neighbors.begin();
						iclique != _neighbors.end(); ++iclique)
				(**iclique).set_number(number, pass);
		};

		const string label () const {
			ostringstream ostream;
			ostream << (NodeSet)*this;
			return(ostream.str());
		};
	};

	typedef list<_Clique> _CliqueList;
	_CliqueList _cliques;

public:
	class Clique: protected _CliqueList::iterator {
		friend class Node;
		friend class Graph;

	private:

		// this is useful for separator cliques
		// it returns the neighboring cliques, assuming there are exactly two
		CliquePair separated_cliques () const {
			CliqueSet::const_iterator iclique = neighbors().begin();
			Clique clique1 = *iclique;
			Clique clique2 = *++iclique;
			return(CliquePair(clique1, clique2));
		};
		// returns the 'other' clique of a separator
		Clique other_clique (Clique clique) const {
			CliquePair cpair = separated_cliques();
			return(cpair.first == clique ? cpair.second : cpair.first);
		};

		void collect (unsigned long pass = 0) {
			if(!pass) pass = ++Graph::pass;
			if((**this).pass == pass) return;
			(**this).pass = pass;
			for(CliqueSet::iterator isep = neighbors().begin();
						isep != neighbors().end(); ++isep) {
				Clique other = isep->other_clique(*this);
				if(other->pass < pass) {
					other.collect(pass);
					(**isep).collect(*other);
					(**this) <<= (**isep);
#ifdef DEBUG
					cout << "message from " << other.label() << " to "
								<< this->label() << endl;
#endif
				};
			};
		};
		void distribute (unsigned long pass = 0) {
			if(!pass) pass = ++Graph::pass;
			if((**this).pass == pass) return;
			(**this).pass = pass;
			for(CliqueSet::iterator isep = neighbors().begin();
						isep != neighbors().end(); ++isep) {
				Clique other = isep->other_clique(*this);
				if(other->pass < pass) {
					(**isep).distribute(**this);
					(*other) <<= (**isep);
					other.distribute(pass);
#ifdef DEBUG
					cout << "message from " << this->label() << " to "
								<< other.label() << endl;
#endif
				};
			};
		};

	public:
		Clique (_CliqueList::iterator it): _CliqueList::iterator(it) {};
		
		bool operator == (Clique clique) const
			{ return(&**this == &*clique); };
		bool operator != (Clique clique) const
			{ return(&**this != &*clique); };
		bool operator < (Clique clique) const
			{ return(&**this < &*clique); };

		bool connected_with (Clique clique, unsigned long pass = 0) const;
		bool operator >> (Clique clique) const
			{ return(connected_with(clique)); };

		const NodeSet& nodes () const { return(**this); };
		const CliqueSet& neighbors () const { return((**this)._neighbors); };

		Clique& operator ++ () {
			_CliqueList::iterator::operator++();
			return(*this);
		};

		const string label () const { return((**this).label()); };
		friend ostream& operator << (ostream& os, const Clique& clique);
		ostream& out_matlab (ostream& os) const
			{ return((**this).out_matlab(os)); };
	};

private:
	// structure to sort separators for a maximum spanning tree
	struct SeparatorCmp {
		bool operator () (Clique a, Clique b) const {
			if(b->size() < a->size()) return(true);
			if(a->size() < b->size()) return(false);
			if(a->union_size < b->union_size) return(true);
			if(b->union_size < a->union_size) return(false);
			return(&*a < &*b);
		};
	};

private:
	_NodeList::iterator _first_node, _last_node;
	_CliqueList::iterator _first_clique, _last_clique;

	Clique insert (const _Clique& _clique);

	void init_node_neighbors ();
	void init_clique_neighbors ();

public:
	Graph () {
		_first_node = _last_node = _nodes.end();
		_first_clique = _last_clique = _cliques.end();
	};

	Node first_node () const { return(_first_node); };
	Node last_node () const { return(_last_node); };
	Clique first_clique () const { return(_first_clique); };
	Clique last_clique () const { return(_last_clique); };

	Node node (char *label = NULL) {
		_nodes.push_front(_Node(label));
		return(_first_node = _nodes.begin());
	};
	Clique clique (NodeSet ns, const double *p) {
		return(insert(_Clique(ns, p)));
	};
	Clique clique (int n, const Node *x, const double *p) {
		// Segfault occurs here
//		cout << "[cli "; cout.flush();
//		_Clique tmp(n, x, p);
//		cout << "clii] "; cout.flush();
		return(insert(_Clique(n, x, p)));
	};
	CLIQUE_FUNCS

	void erase (Node node);
	void erase (Clique clique);

	// returns 0 if the clique size exceeds max somewhere in the process,
	// otherwise the tree width is returned
	int tree_width (int max = 100);
	int triangulate (int max = 32);

	// now returns the partition sum Z
	double marginalize ();

	friend ostream& operator << (ostream& os, const Graph& graph);
};

typedef Graph::Node Node;
typedef Graph::NodeSet NodeSet;
typedef Graph::Clique Clique;
typedef Graph::CliqueSet CliqueSet;

#endif

