//
// LoopGraph.hh
// 
// Made by Vicen Gomez
// Login   <vicen@localhost.localdomain>
// 
// Started on  Tue May 16 10:54:06 2006 Vicen Gomez
// Last update Sat Sep  2 19:34:11 2006 Vicen
//

#ifndef   	LOOPS_HH_
# define   	LOOPS_HH_

#include "fgraph.hh"
#include <boost/dynamic_bitset.hpp>

typedef	boost::dynamic_bitset<unsigned long> BitSet;
typedef unsigned short LVertex;
typedef std::vector<LVertex> VectorLVertex;
typedef std::vector<std::vector<Real> > Mu;   
typedef std::pair<Real, BitSet> LoopTerm;

using namespace boost;


class LoopGraph : public FGraph
{
public:

  LoopGraph(std::string filename, TBPVariant bp_variant=BP_SEQFIX, int prec=10, int ran=0, int verbosity=0)
		: FGraph(filename, bp_variant, prec, ran, verbosity) { }

  virtual ~LoopGraph();

  bool doLoopSeries(bool output = false);

  static BGraph gcore;                // 2-core of the original graph
	static Mu m_mus;                    // precalculated mu terms

  typedef set<BitSet> MapGLoop;
  typedef set<BitSet>::iterator MapGLoopIt;

	vector<LoopTerm> m_contrib;
protected:
  VectorLVertex m_core2g;
	VectorLVertex m_g2core;

	void buildCoreMap();
  void preprocessGraph();
  void computeMusFactorGraph();
  void computeMusForneyGraph();

	uint m_S;                         // number of single loops
  uint m_max_length;                // max length of all loops
	Real m_W;                         // lower bound abs r(C)
	Real m_V;                         // upper bound abs r(C)

	double m_cpu_tlsbp;

	vector<uint> m_node_pos;

	bool m_end;
	MapGLoop m_sloops;                // all simple loops 
	MapGLoop m_new_sloops;            // simple loops of one iteration
	MapGLoop m_gloops;                // all non simple loops
	MapGLoop m_new_gloops;            // non simple loops of one iteration

	BitSet m_marked;

	int m_even_loops;
  int m_zero_loops;

	Real m_logZ_tlsbp;
	Real m_sumR;
	Real m_sum_single;

	Real getR(const BitSet &subgraph);
	Real getRForney(const BitSet &subgraph);

	uint m_mergings;
public:

  class SFactor {
	public:
		uint i, j, k, l;                // bits marked in the subgraph
		SFactor(const uint &_i, const uint &_j, const uint &_k, const uint &_l)	{ i = _i; j = _j; k = _k; l = _l; }
		SFactor()	{ i = j = k = l = 0; }

	};
	vector<SFactor> m_sfactors;

	Real big_terms; 
	uint getSLoopSize() const { return m_sloops.size(); }
	uint getGLoopSize() const { return m_gloops.size(); }
	uint getNoLoopSize() const { return 0; } //m_noloops.size(); }
	uint getMergings() const { return m_mergings; }
	double getCpuTLSBP() const { return m_cpu_tlsbp; }
	Real getLogZTLSBP() const { return m_logZ_tlsbp + Log(m_sumR); }
	Real getLogZSingle() const { return m_logZ_tlsbp + Log(m_sum_single); }
	Real getLogZExact() const { return Real(m_logZ_exact); }
	void loop2Dot(const BitSet &loop);

	// for the COMBINATORIAL recursive search
public:
  class PFactor {
		public:
		uint i, j;                      // adjacent vertices
		Real muf;                  // loop term
		BitSet subgraph;                // nodes of the graph
		BitSet adjacent;                // adjacent pfactors (allowed)

		PFactor(const uint &_i, const uint &_j, const Real &_muf, const BitSet &_v, const BitSet &_a)
		{ i = _i; j = _j; muf = _muf; subgraph = _v; adjacent = _a; }

	};

	vector<PFactor> m_pfactors;
  BitSet m_subgraph;                     // bitset of NODES (pair-wise case)
	BitSet m_allowed;                      // m_allowed[i] == false if pfactor i has been used
	uint m_first_var;
	uint m_nloops;
	uint m_nloops_ok;
	uint m_length;
	uint m_deadends;
	vector<uint> m_degrees;
	Real m_mu;

	// for efficient merging of disconnected loops
	public:
  class Disconnected {
		public:
		  BitSet loop;
			Real term;
			uint nmiss;
			uint nterms;

			Disconnected(const uint &N) : loop(N, 0) {
				term = 1;
				nmiss = 0;
				nterms = 0;
			}
	};

};

std::ostream& operator<< (std::ostream& os, const LoopGraph::SFactor& p);
std::ostream& operator<< (std::ostream& os, const LoopGraph::PFactor& p);
std::ostream& operator<< (std::ostream& os, const vector<BitSet>& sg);


#endif	    /* !LOOPGRAPH_HH_ */
