#include "pindex.hh"
#include "stat.hh"

#include "fgraph.hh"

#include <iostream>
#include <vector>
#include <algorithm>
#include <fstream>
#include <iomanip>

#include <dai/alldai.h>  // Include main libDAI header file
#include <dai/bp.h>
#include <dai/decmap.h>

#include <dai/jtree.h>
#include <dai/factorgraph.h>

#include <boost/dynamic_bitset.hpp>
#include <boost/program_options.hpp>

typedef	boost::dynamic_bitset<unsigned long> Bitset;
typedef std::vector<unsigned char> Degrees;

using namespace std;
namespace po = boost::program_options;

namespace dai {

  std::vector<Real> calcDists( const vector<Factor> &x, const vector<Factor> &y ) {
      vector<Real> errs;
      errs.reserve( x.size() );
      DAI_ASSERT( x.size() == y.size() );
      for( size_t i = 0; i < x.size(); i++ )
          errs.push_back( dist( x[i], y[i], DISTTV ) );
      return errs;
  }

  /// Wrapper class for DAI approximate inference algorithms
  class TestDAI {
    protected:
        /// Stores a pointer to an InfAlg object, managed by this class
        InfAlg          *obj;
        /// Stores the name of the InfAlg algorithm
        string          name;
        /// Stores the total variation distances of the variable marginals
        vector<Real>    varErr;
        /// Stores the total variation distances of the factor marginals
        vector<Real>    facErr;

    public:
        /// Stores the variable marginals
        vector<Factor>  varMarginals;
        /// Stores the factor marginals
        vector<Factor>  facMarginals;
        /// Stores all marginals
        vector<Factor>  allMarginals;
        /// Stores the logarithm of the partition sum
        Real            logZ;
        /// Stores the maximum difference in the last iteration
        Real            maxdiff;
        /// Stores the computation time (in seconds)
        double          time;
        /// Stores the number of iterations needed
        size_t          iters;
        /// Does the InfAlg support logZ()?
        bool            has_logZ;
        /// Does the InfAlg support maxDiff()?
        bool            has_maxdiff;
        /// Does the InfAlg support Iterations()?
        bool            has_iters;

        /// Construct from factor graph \a fg, name \a _name, and set of properties \a opts
        TestDAI( const FactorGraph &fg, const string &_name, const PropertySet &opts ) : obj(NULL), name(_name), varErr(), facErr(), varMarginals(), facMarginals(), allMarginals(), logZ(0.0), maxdiff(0.0), time(0), iters(0U), has_logZ(false), has_maxdiff(false), has_iters(false) {
            double tic = toc();

            if( name == "LDPC" ) {
                // special case: simulating a Low Density Parity Check code
                Real zero[2] = {1.0, 0.0};
                for( size_t i = 0; i < fg.nrVars(); i++ )
                    varMarginals.push_back( Factor(fg.var(i), zero) );
                allMarginals = varMarginals;
                logZ = 0.0;
                maxdiff = 0.0;
                iters = 1;
                has_logZ = false;
                has_maxdiff = false;
                has_iters = false;
            } else
                // create a corresponding InfAlg object
                obj = newInfAlg( name, fg, opts );

            // Add the time needed to create the object
            time += toc() - tic;
        }

        /// Destructor
        ~TestDAI() {
            if( obj != NULL )
                delete obj;
        }

        /// Identify
        string identify() const {
            if( obj != NULL )
                return obj->identify();
            else
                return "NULL";
        }

        /// Run the algorithm and store its results
        void doDAI() {
            double tic = toc();
            if( obj != NULL ) {
                // Initialize
                obj->init();
                // Run
                obj->run();
                // Record the time
                time += toc() - tic;

                // Store logarithm of the partition sum (if supported)
                try {
                    logZ = obj->logZ();
                    has_logZ = true;
                } catch( Exception &e ) {
                    if( e.getCode() == Exception::NOT_IMPLEMENTED )
                        has_logZ = false;
                    else
                        throw;
                }

                // Store maximum difference encountered in last iteration (if supported)
                try {
                    maxdiff = obj->maxDiff();
                    has_maxdiff = true;
                } catch( Exception &e ) {
                    if( e.getCode() == Exception::NOT_IMPLEMENTED )
                        has_maxdiff = false;
                    else
                        throw;
                }

                // Store number of iterations needed (if supported)
                try {
                    iters = obj->Iterations();
                    has_iters = true;
                } catch( Exception &e ) {
                    if( e.getCode() == Exception::NOT_IMPLEMENTED )
                        has_iters = false;
                    else
                        throw;
                }

                // Store variable marginals
                varMarginals.clear();
                for( size_t i = 0; i < obj->fg().nrVars(); i++ )
                    varMarginals.push_back( obj->beliefV( i ) );

                // Store factor marginals
                facMarginals.clear();
                for( size_t I = 0; I < obj->fg().nrFactors(); I++ )
                    try {
                        facMarginals.push_back( obj->beliefF( I ) );
                    } catch( Exception &e ) {
                        if( e.getCode() == Exception::BELIEF_NOT_AVAILABLE )
                            facMarginals.push_back( Factor( obj->fg().factor(I).vars(), INFINITY ) );
                        else
                            throw;
                    }

                // Store all marginals calculated by the method
                allMarginals = obj->beliefs();
            };
        }

        /// Calculate total variation distance of variable and factor marginals with respect to those in \a varMargs and \a facMargs
        void calcErrors( const vector<Factor>& varMargs, const vector<Factor>& facMargs ) {
            varErr = calcDists( varMarginals, varMargs );
            facErr = calcDists( facMarginals, facMargs );
        }

        /// Return maximum variable error
        Real maxVarErr() {
            return( *max_element( varErr.begin(), varErr.end() ) );
        }

        /// Return average variable error
        Real avgVarErr() {
            return( accumulate( varErr.begin(), varErr.end(), 0.0 ) / varErr.size() );
        }

        /// Return maximum factor error
        Real maxFacErr() {
            return( *max_element( facErr.begin(), facErr.end() ) );
        }

        /// Return average factor error
        Real avgFacErr() {
            return( accumulate( facErr.begin(), facErr.end(), 0.0 ) / facErr.size() );
        }
      };
};



int main(int argc, char* argv[])
{
	int ret = -1;

  try {
    string filename;
    string aliases;
    vector<string> methods;
    double tol;
    size_t all_comb = 0;
    size_t maxiter;
    size_t verbose;
 
    int prec = 15, R = 0;

    po::options_description opts_required("Required options", 100);
    opts_required.add_options()
      ("filename", po::value< string >(&filename), "Filename of FactorGraph")
      ("methods", po::value< vector<string> >(&methods)->multitoken(), "AI methods to test")
      ("all_comb", po::value< size_t >(&all_comb), "if 0 computes Z_2 corrections. If 1, all combinations of triplets (default 0)")
      ;

    po::options_description opts_optional("Allowed options", 100);
    opts_optional.add_options()
      ("help", "produce help message")
      ("aliases", po::value< string >(&aliases), "Filename for aliases")
      ("tol", po::value< double >(&tol), "Override tolerance")
      ("maxiter", po::value< size_t >(&maxiter), "Override maximum number of iterations")
      ("verbose", po::value< size_t >(&verbose), "Override verbosity")
      ("R", po::value< int >(&R), "random seed. If R!=0, BP messages are initialized randomly")
      ;

    po::options_description cmdline_options(100);
    cmdline_options.add(opts_required).add(opts_optional);

    po::variables_map vm;
    po::store(po::parse_command_line(argc, argv, cmdline_options), vm);      	    
    po::notify(vm);

    if( vm.count("help") || !(vm.count("filename") && vm.count("methods")) ) {
      cout << "Reads factorgraph <filename.fg> and performs the approximate" << endl;
      cout << "inference algorithms <method*>, reporting clocks, max and average" << endl;
      cout << "error and relative logZ error (comparing with the results of" << endl;
      cout << "<method0>, the base method, for which one can use JTREE_HUGIN)." << endl << endl;
      cout << opts_required << opts_optional << endl;
      return 1;
    }

    Real bptol = tol;
    uint bpmaxiter = 1000;
    int variant = 2, tries = 0;

    int max_states = -1;
		double cpu_jtree = 0;

    // Create loop graph
    FGraph graph(filename, (TBPVariant)variant, prec, R, verbose);

		// Run BP (try the following schedulings)
    bool conv = false;
    while (!conv && (tries < 5) ) {

			if (verbose) {
        switch (variant) {
          case 1: cerr << "BP_SEQFIXED" << endl; break;  
          case 2: cerr << "BP_SEQRANDOM" << endl; break;  
          case 3: cerr << "BP_SEQMAXRES" << endl; break;  
          case 4: cerr << "BP_PAR" << endl; break;  
        }
			}
      graph.setTBPVariant( (TBPVariant) variant );
      if (graph.doBP(bptol, bpmaxiter, false)) {
        conv = true;
			}
      else {
        if (verbose) cerr << graph.m_maxdif << endl;
        variant = (variant+1)%4;
        tries++;
      }
    }

    if (conv) {

			// compute the Pfaffian Series in the Forney representation
			// toForneyStyle already sets the result of BP in the Forney graph so we don't need to run
			// BP in the Forney graph which takes more iterations (and sometimes worse convergence
			graph.toForneyStyle();
			graph.doPfaffianSeries( all_comb );


			if (Abs(graph.getLogZBPForney() - graph.getLogZBP()) > Real(1e-10)) {
				cerr << "LogZ differs between Forney (" << graph.getLogZBPForney() << ") and Factor (" << graph.getLogZBP();
				cerr << ") diff = " << Abs(graph.getLogZBPForney() - graph.getLogZBP()) << endl;
			}

      // Read aliases
      map<string,string> Aliases;
      if( !aliases.empty() ) {
        ifstream infile;
        infile.open (aliases.c_str());
        if (infile.is_open()) {
          while( true ) {
            string line;
            getline(infile,line);
            if( infile.fail() )
              break;
            if( (!line.empty()) && (line[0] != '#') ) {
              string::size_type pos = line.find(':',0);
              if( pos == string::npos )
                throw "Invalid alias";
              else {
                string::size_type posl = line.substr(0, pos).find_last_not_of(" \t");
                string key = line.substr(0, posl + 1);
                string::size_type posr = line.substr(pos + 1, line.length()).find_first_not_of(" \t");
                string val = line.substr(pos + 1 + posr, line.length());
                Aliases[key] = val;
              }
            }
          }
          infile.close();
        } else
          throw "Error opening aliases file";
      } 

      dai::FactorGraph fg;
      fg.ReadFromFile(filename.c_str());
			boundTreewidth(fg,&dai::eliminationCost_MinFill,max_states);

      vector<dai::Factor> q0;
      vector<dai::Factor> v0;
      double logZ0 = 0.0;
      double init_clock = 0;

      // For each method...
      for( size_t m = 0; m < methods.size(); m++ ) {

        // Parse method
        pair<string, dai::PropertySet> meth = dai::parseNameProperties( methods[m], Aliases );

        // Construct object for running the method
        dai::TestDAI testdai(fg, meth.first, meth.second );

        if( meth.first == string("JTREE") )
          init_clock = dai::toc();

        // Run the method
        testdai.doDAI();
        if( meth.first == string("JTREE") ) cpu_jtree = ( dai::toc() - init_clock );

				// Joris considers that the first method is the exact
				if ( m == 0 ) {        
          v0 = testdai.facMarginals;
          q0 = testdai.varMarginals;
					logZ0 = testdai.logZ;
					if( meth.first == string("JTREE") ) {
						
						// I consider the JTREE as the only exact method          
						for (uint i=0; i<q0.size(); i++) {
							VProb pi(Real(2.));
							//for (uint v=0; v<q0[i].stateSpace(); v++) {
             for (uint v=0; v<q0[i].nrStates(); v++) {
								pi[v] = Real(q0[i][v]);
							}
							graph.setExactMarginal(i, pi);
						}
						graph.m_logZ_exact = logZ0;
					}
				}
				testdai.calcErrors(q0, v0);
      }

			// output results cout
			graph.outputRes();

			// output results cerr
			Real abs_logZ = Abs(graph.m_logZ_exact);
			ret = ((Abs(graph.getLogZPfaffian() - abs_logZ) / abs_logZ) < Real(1e-14)) ? 0 : 1;

			cout << setprecision(14) << scientific;

			cout << graph.getBPVariant() << " ";	// 1 
			/////////////
			cout << graph.getCpuBP() << " ";	// 2
			cout << graph.getLogZBP() << " "; // 4
			/////////////
			cout << max_states << " "; // 5
			cout << cpu_jtree << " "; // 6
			cout << graph.getLogZExact() << " "; // 8
			/////////////
			cout << graph.getCpuPfaffian1() << " "; // 9
			cout << graph.getLogZPfaffian1() << " "; // 11
			/////////////
			cout << graph.getCpuPfaffian() << " "; // 12
			cout << graph.getLogZPfaffian() << " "; // 14
			/////////////
			cout << "0 "; // 15
			cout << "0 "; // 16

		} // end convergence 
		else {

			cout << graph.getBPVariant() << " "; 
			/////////////
			cout << " -1 ";
			cout << " -1 ";
			/////////////
			cout << max_states << " ";
			cout << cpu_jtree << " ";
			cout << graph.getLogZExact() << " ";
			/////////////
			cout << " -1 ";
			cout << " -1 ";
			/////////////
			cout << " -1 ";
			cout << " -1 ";
			/////////////
			cout << " -1 ";
			cout << " -1\t";

			return ret;
		}
  } catch(const char *e) {
    cerr << "Exception: " << e << endl;
    return 1;
  } catch(std::exception& e) {
    cerr << "Exception: " << e.what() << endl;
    return 1;
  }
  catch(...) { 
    cerr << "Exception of unknown type!" << endl;
  }

  return ret;
}


//				cout << graph.getSLoopSize() << " ";
//				cout << graph.getGLoopSize() << " ";
//				cout << graph.getNoLoopSize() << " ";
//				cout << graph.getMergings() << " ";
//				cout << graph.getCpuTLSBP() << " ";
//				cout << toSeconds( graph.getCpuTLSBP() ) << " ";
//				cout << graph.m_nloops_ok << " ";
//				cout << graph.m_nloops << " ";

